from __future__ import annotations

import argparse
import csv
import subprocess
import sys
import typing

from ldaptool import decode, search
from ldaptool._utils.ldap import Result, SizeLimitExceeded


class _Context:
    def __init__(self) -> None:
        parser = argparse.ArgumentParser()
        arguments_p = search.Arguments.add_to_parser(parser)
        args = parser.parse_args()
        try:
            self.config = search.Config.load()
        except Exception as e:
            raise SystemExit(f"config error: {e}")
        self.arguments = arguments_p.from_args(args)

    def run(self) -> None:
        # starting the search sets the base we want to print
        search_iterator = search.search(config=self.config, arguments=self.arguments)
        self._run_with_filters(search_iterator)

    def _run_with_filters(self, search_iterator: typing.Iterable[Result]) -> None:
        output: typing.IO[str] = sys.stdout
        procs: list[subprocess.Popen[str]] = []

        def add_filter(cmd: list[str]) -> None:
            nonlocal output
            proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=output, text=True)
            procs.append(proc)
            if output != sys.stdout:
                output.close()
            assert proc.stdin
            output = proc.stdin

        try:
            if self.arguments.table:
                add_filter(["csvlook"])
            if self.arguments.sort:
                add_filter(["csvsort", "--blanks"])
            self._run_search(search_iterator, stream=output)
        finally:
            if procs:
                output.close()
            for proc in reversed(procs):
                proc.wait()

    def _run_search(self, search_iterator: typing.Iterable[Result], *, stream: typing.IO[str]) -> None:
        decoder = decode.Decoder(arguments=self.arguments)

        num_responses = 0
        num_entries = 0

        ldif_output = not (self.arguments.csv or self.arguments.json or self.arguments.human)

        if ldif_output:
            print("# extended LDIF")
            print("#")
            print("# LDAPv3")
            print(f"# base <{self.arguments.base}> with scope subtree")
            print(f"# filter: {self.arguments.filter}")
            if self.arguments.attributes:
                print(f"# requesting: {' '.join(self.arguments.attributes)}")
            else:
                print("# requesting: ALL")
            print("#")
            print()

        if self.arguments.csv:
            csv_out = csv.DictWriter(
                stream,
                fieldnames=self.arguments.columns,
                lineterminator="\n",
                extrasaction="ignore",
            )
            csv_out.writeheader()
            # dicts contain data by lower case key
            csv_out.fieldnames = [col.lower() for col in self.arguments.columns]

        try:
            for dn, entry in search_iterator:
                num_responses += 1
                if dn is None:
                    if not self.arguments.csv:
                        print("# search reference")
                        for ref in entry:
                            assert isinstance(ref, str)
                            print(f"ref: {ref}")
                        print()
                    continue
                # normal entry
                assert not isinstance(entry, list)
                num_entries += 1
                obj = decoder.read(dn=dn, entry=entry)
                if self.arguments.csv:
                    csv_out.writerow(decoder.human(dn=dn, entry=obj))
                else:
                    decoder.emit(dn=dn, entry=obj)
        except SizeLimitExceeded as e:
            raise SystemExit(f"Error: {e}")

        if ldif_output:
            print(f"# numResponses: {num_responses}")
            print(f"# numEntries: {num_entries}")


def main() -> None:
    ctx = _Context()
    ctx.run()
